#include "read.h"

#include <ctype.h>
#include <inttypes.h>
#include <limits.h>
#include <string.h>

struct ReadState {
    const char *head;
    size_t left;
    size_t off;
    size_t line;
    size_t col;

    size_t backquote_level;
};

#define EOS -1

static int popc(struct ReadState *state) {
    if (!state->left) {
        return EOS;
    }
    ++state->off;
    --state->left;
    int c = *(state->head++);
    if (c == '\n') {
        ++state->line;
        state->off = 0;
    } else {
        ++state->col;
    }
    return c;
}

static int peekc(struct ReadState *state) {
    if (!state->left) {
        return EOS;
    }
    return *state->head;
}

static inline void _internal_read_error(struct ReadState *state, size_t len,
                                        LispVal *desc) {
    // TODO format better
    LispVal *args = make_list(
        4, make_lisp_integer(state->line), make_lisp_integer(state->col),
        make_lisp_string(state->head, len, false, false), desc);
    lisp_ref(args);
    Fthrow(Qread_error, args);
    UNREF_INPLACE(args);
}
#define READ_ERROR(state, len, ...) \
    _internal_read_error(state, len, sprintf_lisp(__VA_ARGS__))
#define EOF_ERROR(state) READ_ERROR(state, 1, "unexpected end of file")

#define SKIP_WHILE(cond, state) \
    while (cond) {              \
        popc(state);            \
    }
#define SKIP_WHITESPACE(state) SKIP_WHILE(isblank(peekc(state)), state)

static bool is_symbol_end(int c) {
    return c == EOS || isspace(c) || c == '(' || c == ')' || c == '['
           || c == ']';
}

static LispVal *read_internal(struct ReadState *state);

static LispVal *read_list(struct ReadState *state) {
    popc(state); // open (
    LispVal *list = Qnil;
    LispVal *end = list;
    SKIP_WHITESPACE(state);
    int c;
    while ((c = peekc(state)) != ')') {
        if (c == EOS) {
            EOF_ERROR(state);
            UNREF_INPLACE(list);
            return Qnil;
        }
        LispVal *elt = read_internal(state);
        if (NILP(list)) {
            list = Fpair(elt, Qnil);
            end = list;
        } else {
            LispVal *new_end = Fpair(elt, Qnil);
            Fsettail(end, new_end);
            end = new_end;
        }
        SKIP_WHITESPACE(state);
    }
    popc(state); // close )
    return list;
}

static LispVal *read_vector(struct ReadState *state) {
    popc(state); // open [
    LispVal **values = NULL;
    size_t values_len = 0;
    SKIP_WHITESPACE(state);
    int c;
    while ((c = peekc(state)) != ']') {
        if (c == EOS) {
            EOF_ERROR(state);
            for (size_t i = 0; i < values_len; ++i) {
                lisp_unref(values[i]);
            }
            lisp_free(values);
            return Qnil;
        }
        LispVal *elt = lisp_ref(read_internal(state));
        values = lisp_realloc(values, sizeof(LispVal *) * ++values_len);
        values[values_len - 1] = elt;
        SKIP_WHITESPACE(state);
    }
    popc(state); // close ]
    return make_lisp_vector(values, values_len);
}

static LispVal *read_string(struct ReadState *state) {
    popc(state); // open "
    bool backslash = false;
    int c;
    char *str = lisp_malloc(1);
    size_t str_len = 0;
    while (backslash || (c = peekc(state)) != '"') {
        if (c == EOS) {
            lisp_free(str);
            EOF_ERROR(state);
            return Qnil;
        }
        popc(state);
        if (!backslash && c == '\\') {
            backslash = true;
        } else if (backslash && c == '\n') {
            backslash = false;
        } else {
            str = lisp_realloc(str, ++str_len + 1);
            int to_add = c;
            if (backslash) {
                switch (c) {
                case 'n':
                    to_add = '\n';
                    break;
                case 't':
                    to_add = '\t';
                    break;
                case 'r':
                    to_add = '\r';
                    break;
                case '0':
                    to_add = '\0';
                    break;
                case '"':
                    to_add = '"';
                    break;
                default:
                    // TODO make this point at the correct thing
                    READ_ERROR(state, 1, "unknown escape sequence");
                    lisp_free(str);
                    return Qnil;
                }
            }
            backslash = false;
            str[str_len - 1] = to_add;
        }
    }
    str[str_len] = '\n';
    popc(state); // close "
    return make_lisp_string(str, str_len, true, false);
}

static LispVal *read_character(struct ReadState *state) {
    static struct {
        const char *name;
        size_t len;
        char value;
    } LOOKUP_TABLE[] = {
        {"null", 4, '\0'}, {"\\0", 2, '\0'}, {"newline", 7, '\n'},
        {"\\n", 2, '\n'},  {"tab", 3, '\t'}, {"\\t", 2, '\t'},
    };
#define LOOKUP_TABLE_SIZE (sizeof(LOOKUP_TABLE) / sizeof(LOOKUP_TABLE[0]))
    struct ReadState start_state = *state;
    popc(state); // #
    const char *start = state->head;
    while (!is_symbol_end(peekc(state))) {
        popc(state);
    }
    size_t len = state->head - start;
    for (size_t i = 0; i < LOOKUP_TABLE_SIZE; ++i) {
        if (len == LOOKUP_TABLE[i].len
            && strings_equal_nocase(start, LOOKUP_TABLE[i].name, len)) {
            return make_lisp_integer(LOOKUP_TABLE[i].value);
        }
    }
    READ_ERROR(&start_state, len, "unknown character liternal: %*s", (int) len,
               start);
    return Qnil;
#undef LOOKUP_TABLE_SIZE
}

#define INVALID_BASE -2
#define ANY_BASE     -1

static bool is_base_char(int base, int c) {
    c = tolower(c);
    if (base == ANY_BASE) {
        base = 10;
    }
    switch (base) {
    case 16:
        if (c >= 'a' || c <= 'f') {
            return true;
        }
    case 10:
    case 8:
    case 2:
        return c >= '0' && c <= '0' + base - 1;
    default:
        return false;
    }
}

static int parse_base(size_t left, const char *c) {
    if (left >= 2 && *c == '2' && *(c + 1) == '#') {
        return 2;
    } else if (left >= 2 && *c == '8' && *(c + 1) == '#') {
        return 8;
    } else if (left >= 3 && *c == '1' && *(c + 1) == '0' && *(c + 2) == '#') {
        return 10;
    } else if (left >= 3 && *c == '1' && *(c + 1) == '6' && *(c + 2) == '#') {
        return 16;
    } else {
        return INVALID_BASE;
    }
}

static LispVal *read_symbol(struct ReadState *state) {
    const char *start = state->head;
    while (!is_symbol_end(peekc(state))) {
        popc(state);
    }
    return intern(start, state->head - start, false);
}

static LispVal *read_number_or_symbol(struct ReadState *state, int base) {
    size_t size_t_len = ((sizeof(size_t) * 8) / 3) + 1;
    char fmt_buf[3 + size_t_len + 1];
    struct ReadState start_state = *state;
    bool has_decimal = false;
    const char *number_start = state->head;
    const char *exp_start = NULL;
    int c;
    while (!is_symbol_end(c = peekc(state))) {
        popc(state);
        if (c == '#') {
            if (base != ANY_BASE) {
                goto change_to_symbol;
            }
            base = parse_base(start_state.left, start_state.head);
            if (base == INVALID_BASE) {
                goto change_to_symbol;
            }
            number_start = state->head;
        } else if (c == '.') {
            if (base != ANY_BASE || has_decimal || exp_start) {
                goto change_to_symbol;
            }
            has_decimal = true;
        } else if ((c == '-' || c == '+')) {
            if (state->head - 1 != number_start
                && state->head - 1 != exp_start) {
                goto change_to_symbol;
            }
            // fallthrough
        } else if (!is_base_char(base, c)) {
            if ((c == 'e' || c == 'E') && !exp_start && base == ANY_BASE) {
                exp_start = state->head;
            } else {
                goto change_to_symbol;
            }
        }
    }
    size_t len = state->head - number_start;
    // ceil(# bytes in size_t / 3)
    // This works because log10(2^n) is O(n) for k=3
    if (has_decimal || exp_start) {
        if (exp_start == state->head) {
            goto change_to_symbol;
        }
        snprintf(fmt_buf, sizeof(fmt_buf), "%%%zuLf", len);
        long double value;
        sscanf(number_start, fmt_buf, &value);
        return make_lisp_float(value);
        // float
    } else {
        snprintf(fmt_buf, sizeof(fmt_buf), "%%%zujd", len);
        intmax_t value;
        sscanf(number_start, fmt_buf, &value);
        return make_lisp_integer(value);
        // int
    }
    abort();
change_to_symbol:
    *state = start_state;
    return read_symbol(state);
}

static LispVal *read_internal(struct ReadState *state) {
    SKIP_WHILE(isspace(peekc(state)), state);
    int c = peekc(state);
    // comment
    while (c == ';') {
        SKIP_WHILE(peekc(state) != '\n', state);
        SKIP_WHITESPACE(state);
        c = peekc(state);
    }
    switch (c) {
        // list
    case EOS:
        return NULL;
    case ')':
        READ_ERROR(state, 1, "unmatched \")\"");
        return Qnil;
    case ']':
        READ_ERROR(state, 1, "unmatched \"]\"");
        return Qnil;
    case '(':
        return read_list(state);
        // string
    case '"':
        return read_string(state);
        // vector
    case '[':
        return read_vector(state);
        // quoted form
    case '\'': {
        popc(state); // '
        LispVal *tail = read_internal(state);
        return Fpair(Qquote, Fpair(tail, Qnil));
    }
        // backquote
    case '`': {
        popc(state); // `
        ++state->backquote_level;
        LispVal *tail = read_internal(state);
        --state->backquote_level;
        return Fpair(Qbackquote, Fpair(tail, Qnil));
    }
        // comma
    case ',':
        popc(state); // ,
        if (state->backquote_level) {
            --state->backquote_level;
            LispVal *tail = read_internal(state);
            ++state->backquote_level;
            return Fpair(Qcomma, Fpair(tail, Qnil));
        } else {
            READ_ERROR(state, 1, "comma not inside backquote");
            return Qnil;
        }
        // character literal
    case '#':
        return read_character(state);
        // number
    case '+':
    case '-':
    case '.':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    case '0':
        return read_number_or_symbol(state, ANY_BASE);
    // symbol
    default:
        return read_symbol(state);
    }
}

size_t read_from_buffer(const char *text, size_t length, LispVal **out) {
    struct ReadState state = {
        .head = text,
        .left = length,
        .off = 0,
        .line = 1,
        .col = 0,
        .backquote_level = 0,
    };
    LispVal *res = read_internal(&state);
    if (!res) {
        EOF_ERROR(&state);
        *out = Qnil;
    } else {
        *out = res;
    }
    return state.off;
}
